from typing import Optional, Dict, Literal, List
from enum import Enum
import json

from . import _constants as constants
from ._api import OpenAPI


class BaseAnalysis(OpenAPI):
    def _extract_id(self, data_id: str, table_id: Optional[str] = None):
        data_details = self.get_sample_data_detail(data_id)
        study_id = data_details[constants.ConnectorKeys.STUDY_ID.value]
        sample_id = data_details[constants.ConnectorKeys.SAMPLE_ID.value]
        if table_id is None:
            table_results = []
            submitted_results: Dict[str, Dict[str, dict]] = \
                data_details[constants.ConnectorKeys.MAP_SUBMIT_RESULT.value]
            for _, res in submitted_results.items():
                table_results.extend(
                    res.get(constants.ConnectorKeys.ANNOTATED_DATA.value, {}).keys())
            if len(table_results) == 0:
                raise RuntimeError("Not found any annotated data.")
            if len(table_results) > 1:
                raise RuntimeError(
                    f"Found too many annotated data: {table_results}. "
                    "Please identify an ID by `table_id=...`"
                )
            table_id = table_results[0]
        return study_id, sample_id, table_id

    def _post_analysis(
        self,
        data_id: str,
        title: str,
        params: dict,
        sub_type: str,
        group_type: str,
        table_id: Optional[str] = None,
        **_,
    ):
        study_id, sample_id, table_id = self._extract_id(data_id, table_id=table_id)
        return self.post_openapi_request(
            url=constants.CREATE_ANALYSIS_URL,
            json={
                constants.ConnectorKeys.DATA.value: {
                    constants.ConnectorKeys.TITLE.value: title,
                    constants.ConnectorKeys.PARAMS.value: json.dumps(params),
                    constants.ConnectorKeys.DISPLAY_PARAMS.value: json.dumps({}),
                    constants.ConnectorKeys.SUB_TYPE.value: sub_type,
                    constants.ConnectorKeys.GROUP_TYPE.value: group_type,
                    constants.ConnectorKeys.DESCRIPTION.value: title,
                },
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.DATA_ID.value: data_id,
                constants.ConnectorKeys.TABLE_ID.value: table_id,
            }
        )


class Embeddings(BaseAnalysis):
    def pca(
        self,
        data_id: str,
        normalize_method: Optional[
            Literal[
                constants.NormalizeMethod.RAW,
                constants.NormalizeMethod.LOG1P_NORMALIZE,
                constants.NormalizeMethod.SQRT_NORMALIZE,
            ]
        ] = constants.NormalizeMethod.LOG1P_NORMALIZE,
        title: Optional[str] = None,
        batch_key: Optional[str] = None,
        n_top_genes: Optional[int] = None,
        **kwargs,
    ):
        """
        Principal Component Analysis

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        normalize_method: `Literal["raw", "log1p-normalized", "sqrt-normalized"]`, Default: "log1p-normalized"
            Using raw expression or scaling (log1p, square root) to normalize when running PCA.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        batch_key: `Optional[str]`, Default: None
            Metadata which defines batch effect.
            If provided, using Harmony to remove batch effect.
        n_top_genes: `Optional[int]`, Default: None
            Number of top highly variable genes.
            If provided, running highly variable genes to select genes for PCA.
        """
        if isinstance(normalize_method, Enum):
            normalize_method = normalize_method.value

        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "PCA",
            params={
                "normalize_method": normalize_method,
                "name": title,
                "batch_key": batch_key,
                "n_top_genes": n_top_genes,
                **kwargs,
            },
            sub_type="PCA",
            group_type="embedding",
            **kwargs,
        )

    def scvi(
        self,
        data_id: str,
        n_latents: int = 20,
        title: Optional[str] = None,
        batch_key: Optional[str] = None,
        n_top_genes: Optional[int] = 2000,
        **kwargs,
    ):
        """
        Reducing dimensions with scVI

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        n_latents: `int`, Default: 20
            Number of latents.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        batch_key: `Optional[str]`, Default: None
            Metadata which defines batch effect.
            If provided, using Harmony to remove batch effect.
        n_top_genes: `Optional[int]`, Default: 2000
            Number of top highly variable genes.
            If provided, running highly variable genes to select genes for PCA.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "scVI",
            params={
                "n_latents": n_latents,
                "name": title,
                "batch_key": batch_key,
                "n_top_genes": n_top_genes,
                **kwargs,
            },
            sub_type="scVI",
            group_type="embedding",
            **kwargs,
        )

    def umap(
        self,
        data_id: str,
        embedding_key: str,
        n_neighbors: int = 15,
        local_connectivity: float = 1,
        init: Literal["pca", "spectral", "random"] = "pca",
        deterministic: bool = True,
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        Uniform Manifold Approximation and Projection for Dimension Reduction.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        embedding_key: `str`
            Embedding to run UMAP, should be result of PCA or scVI.
        n_neighbors : Optional[`int`], Default: None
            The number of nearest neighbors.
        local_connectivity : `float`, Default: 1.0
            The local connectivity required -- i.e. the number of nearest
            neighbors that should be assumed to be connected at a local level.
            The higher this value the more connected the manifold becomes
            locally. In practice this should be not more than the local intrinsic
            dimension of the manifold.
        init : `str`, Default: "pca"
            How to initialize the low dimensional embedding. Options are:
            * `"spectral"` : use a spectral embedding of the fuzzy 1-skeleton (non-deterministic).
            * `"random"` : assign initial embedding positions at random.
            * `"pca"` : pca low dimensions of input matrix.
        deterministic : `bool`, Default: True
            Making sure with same initialization, results will remains the same across runs on the same device.
            Deterministic version will converge a lot slower on big datasets (hence lower quality on same setting).
            If you prefer quality/performance, disable `deterministic`
            Note: `"spectral"` initialization is non-deterministic across runs.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "UMAP",
            params={
                "embedding_key": embedding_key,
                "n_neighbors": n_neighbors,
                "local_connectivity": local_connectivity,
                "init": init,
                "deterministic": deterministic,
                "name": title,
                **kwargs,
            },
            sub_type="UMAP",
            group_type="embedding",
            **kwargs,
        )

    def tsne(
        self,
        data_id: str,
        embedding_key: str,
        n_neighbors: int = 90,
        perplexity: float = 30,
        local_connectivity: float = 1,
        init: Literal["pca", "gaussian", "input"] = "pca",
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        FFT-accelerated Interpolation-based t-SNE implementation.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        embedding_key: `str`
            Embedding to run UMAP, should be result of PCA or scVI.
        n_neighbors : Optional[`int`], Default: None
            The number of nearest neighbors.
        perplexity : ``float``, default = ``30``
            The perplexity is related to the number of nearest neighbors that is used in
            other manifold learning algorithms.
            Larger datasets usually require a larger perplexity.
            Consider selecting a value between ``5`` and ``50``.
            Different values can result in significantly different results.
        local_connectivity : `float`, Default: 1.0
            The local connectivity required -- i.e. the number of nearest
            neighbors that should be assumed to be connected at a local level.
            The higher this value the more connected the manifold becomes
            locally. In practice this should be not more than the local intrinsic
            dimension of the manifold.
        init : `str`, Default: "pca"
            How to initialize the low dimensional embedding. Options are:
            * ``"gaussian"`` : init using gaussian noise.
            * ``"input"`` : 2 first row input matrix.
            * ``"pca"`` : pca 2 first row input matrix.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "t-SNE",
            params={
                "embedding_key": embedding_key,
                "n_neighbors": n_neighbors,
                "local_connectivity": local_connectivity,
                "init": init,
                "perplexity": perplexity,
                "name": title,
                **kwargs,
            },
            sub_type="t-SNE",
            group_type="embedding",
            **kwargs,
        )


class Clustering(BaseAnalysis):
    def louvain(
        self,
        data_id: str,
        embedding_key: str,
        resolution: float,
        n_neighbors: int = 90,
        local_connectivity: float = 1,
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        Louvain clustering implementation.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        embedding_key: `str`
            Embedding to run UMAP, should be result of PCA or scVI.
        resolution : ``float``
            Resolution parameter, higher lead to more communities.
        n_neighbors : Optional[`int`], Default: None
            The number of nearest neighbors.
        local_connectivity : `float`, Default: 1.0
            The local connectivity required -- i.e. the number of nearest
            neighbors that should be assumed to be connected at a local level.
            The higher this value the more connected the manifold becomes
            locally. In practice this should be not more than the local intrinsic
            dimension of the manifold.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "Louvain",
            params={
                "embedding_key": embedding_key,
                "resolution": resolution,
                "n_neighbors": n_neighbors,
                "local_connectivity": local_connectivity,
                "name": title,
                **kwargs,
            },
            sub_type="Louvain",
            group_type="clustering",
            **kwargs,
        )

    def kmeans(
        self,
        data_id: str,
        embedding_key: str,
        n_clusters: int,
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        k-means clustering implementation.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        embedding_key: `str`
            Embedding to run UMAP, should be result of PCA or scVI.
        n_clusters: ``int``
            Number of clusters.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "k-means",
            params={
                "embedding_key": embedding_key,
                "n_clusters": n_clusters,
                "name": title,
                **kwargs,
            },
            sub_type="k-means",
            group_type="clustering",
            **kwargs,
        )


class DE(BaseAnalysis):
    def differential_expression_genes(
        self,
        data_id_1: str,
        data_id_2: str,
        group_1_indices: List[int],
        group_2_indices: List[int],
        group_1_name: str = "Group 1",
        group_2_name: str = "Group 2",
        method: Literal["venice", "t-test", "wilcoxon"] = "venice",
        normalize_method: Optional[
            Literal[
                constants.NormalizeMethod.RAW,
                constants.NormalizeMethod.LOG1P_NORMALIZE,
                constants.NormalizeMethod.SQRT_NORMALIZE,
            ]
        ] = constants.NormalizeMethod.LOG1P_NORMALIZE,
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        Spatial Region Segmentation.

        Parameters
        ----------
        data_id_1: `str`
            ID of the first data to run differential expression genes.
        data_id_2: `str`
            ID of the second data to run differential expression genes.
        group_1_indices: `List[int]`
            Cell indices to compare in the first group.
        group_2_indices: `List[int]`
            Cell indices to compare in the second group.
        group_1_name: `str`, Default: "Group 1"
            Name of the first group.
        group_2_name: `str`, Default: "Group 2"
            Name of the second group.
        method: `Literal["venice", "t-test", "wilcoxon"]`, Default: "venice"
            Test statistic method.
        normalize_method: `Literal["raw", "log1p-normalized", "sqrt-normalized"]`, Default: "log1p-normalized"
            Using raw expression or scaling (log1p, square root) for test statistic.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        _, sample_id_1, table_id_1 = self._extract_id(data_id_1)
        _, sample_id_2, table_id_2 = self._extract_id(data_id_2)
        if isinstance(normalize_method, Enum):
            normalize_method = normalize_method.value

        return self._post_analysis(
            data_id=data_id_1,
            title=title if title is not None else "DE genes",
            params={
                "group_1_indices": group_1_indices,
                "group_2_indices": group_2_indices,
                "group_1_name": group_1_name,
                "group_2_name": group_2_name,
                "method": method,
                "normalize_method": normalize_method,
                "sample_id_1": sample_id_1,
                "sample_id_2": sample_id_2,
                "table_id_1": table_id_1,
                "table_id_2": table_id_2,
                "name": title,
                **kwargs,
            },
            sub_type="de_genes",
            group_type="deg",
            **kwargs,
        )


class Prediction(BaseAnalysis):
    def metadata_reference(
        self,
        data_id: str,
        cluster_key: str,
        species: str,
        anno_type: Literal["sub", "major"] = "sub",
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        Spatial Region Segmentation.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        cluster_key: `str`
            Cluster to predict cell types.
        species: `str`
            Species of data.
        anno_type: `Literal["sub", "major"]`, Default: "sub"
            Type of annotation: Major or sub cell types.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "Metadata Reference",
            params={
                "cluster_key": cluster_key,
                "anno_type": anno_type,
                "species": species,
                "name": title,
                **kwargs,
            },
            sub_type="celltype_prediction",
            group_type="prediction",
            **kwargs,
        )


class SpatialAnalysis(BaseAnalysis):
    def region_segmentation(
        self,
        data_id: str,
        radius: float,
        mpp: float,
        species: str,
        resolution: float = 0.5,
        title: Optional[str] = None,
        **kwargs,
    ):
        """
        Spatial Region Segmentation.

        Parameters
        ----------
        data_id: `str`
            ID of a data in sample.
        radius: `float`
            Radius (micrometer) to build spatial graph
        mpp: `float`
            Micrometers per pixel.
        species: `str`
            Species of data.
        resolution : ``float``, Default: 0.5
            Resolution parameter, higher lead to more communities.
        title: `Optional[str]`, Default: None
            Title of analysis, it will be used as name of embeddings.
            If not provided, a default name will be used.
        """
        data_details = self.get_sample_data_detail(data_id)
        transcripts = []
        spots = []
        submitted_results: Dict[str, Dict[str, dict]] = \
            data_details[constants.ConnectorKeys.MAP_SUBMIT_RESULT.value]
        for _, res in submitted_results.items():
            transcripts.extend(res.get(constants.ConnectorKeys.TRANSCRIPTS.value, {}).keys())
            spots.extend(res.get(constants.ConnectorKeys.SPOT.value, {}).keys())
        transcripts_id = transcripts[0] if len(transcripts) > 0 else None
        spots_id = spots[0] if len(spots) > 0 else None

        return self._post_analysis(
            data_id=data_id,
            title=title if title is not None else "Region Segmentation",
            params={
                "radius": radius,
                "mpp": mpp,
                "resolution": resolution,
                "species": species,
                "transcripts_id": transcripts_id,
                "spots_id": spots_id,
                "name": title,
                **kwargs,
            },
            sub_type="segmentation_by_spatial_correlation",
            group_type="svg",
            **kwargs,
        )


class Analysis(BaseAnalysis):
    def __init__(
        self,
        domain: str,
        token: str,
        verify_ssl: bool = False,
    ):
        super().__init__(domain, token, verify_ssl=verify_ssl)
        self.embeddings = Embeddings(domain, token, verify_ssl=verify_ssl)
        self.clustering = Clustering(domain, token, verify_ssl=verify_ssl)
        self.de = DE(domain, token, verify_ssl=verify_ssl)
        self.prediction = Prediction(domain, token, verify_ssl=verify_ssl)
        self.spatial_analysis = SpatialAnalysis(domain, token, verify_ssl=verify_ssl)

    def list_metadata(self, data_id: str):
        study_id, sample_id, table_id = self._extract_id(data_id)
        return self.post_pyapi_request(
            url=constants.GET_ANNOTATED_ELEMENTS,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.TABLE_ID.value: table_id,
                constants.ConnectorKeys.ELEMENT.value: constants.ConnectorKeys.OBS.value,
            },
        )

    def list_embedding(self, data_id: str):
        study_id, sample_id, table_id = self._extract_id(data_id)
        return self.post_pyapi_request(
            url=constants.GET_ANNOTATED_ELEMENTS,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.TABLE_ID.value: table_id,
                constants.ConnectorKeys.ELEMENT.value: constants.ConnectorKeys.OBSM.value,
            },
        )



from typing import List, Dict, Union, Optional
import os
from tqdm import tqdm
from pathlib import Path

from ._api import Connector
from ._api import PyAPI
from ._api import OpenAPI
from ._analysis import Analysis
from ._constants import ConnectorKeys
from ._constants import DefaultGroup
from ._utils import get_chunk_size, format_print


class SpatialXConnector(Connector):
    """
    SpatialX Connector
    Supporting to work with spatial data via notebook.
    """

    def __init__(self, domain: str, token: str, verify_ssl: bool = False):
        """
        Construct parameters for train and query k-nearest neighbors

        Parameters
        ----------
        domain: ``str``
            SpatialX domain
        token: ``str``
            User's token
        verify_ssl: ``bool``, default: False
            Verify SSL or not.
        """
        super().__init__(domain, token, verify_ssl)

        self.__pyapi = PyAPI(domain, token, verify_ssl)
        self.__openapi = OpenAPI(domain, token, verify_ssl)
        self.__analysis = Analysis(domain, token, verify_ssl)

    @property
    def pyapi(self) -> PyAPI:
        return self.__pyapi

    @property
    def openapi(self) -> OpenAPI:
        return self.__openapi

    @property
    def analysis(self) -> Analysis:
        return self.__analysis

    @property
    def info(self):
        """Current user's information"""
        info = self.openapi.info
        return {
            field: info[field]
            for field in ConnectorKeys.INFORMATION_FIELDS.value
        }

    @property
    def groups(self):
        """List all reachable groups of current user in domain server."""
        group_info = self.openapi.groups
        groups = {
            v: k for k, v in group_info[ConnectorKeys.DEFAULT.value].items()
        }
        for group in group_info[ConnectorKeys.GROUPS.value]:
            groups[group["name"]] = group["id"]
        return groups

    @property
    def external_folders(self):
        """List all reachable mounted shared folders of current user from BBrowserX/BioStudio."""
        return {
            folder["name"]: folder["path"]
            for folder in self.openapi.mounts["folders"]
        }

    @property
    def folders(self):
        """List all reachable mounted shared folders of current user in domain server."""
        defaults = {
            folder["name"]: folder["path"]
            for folder in self.openapi.info["default_mount"]["folders"]
        }
        return dict(self.external_folders.items() | defaults.items())

    @property
    def s3(self):
        """List all reachable mounted s3 clouds of current user in domain server."""
        s3_buckets = {}
        for s3 in self.openapi.info["default_mount"]["s3"]:
            name = s3.get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = s3["id"]
            s3_buckets[name] = s3["path"]

        for external_s3 in self.openapi.mounts["s3"]:
            name = external_s3.get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = external_s3["id"]
            if name in s3_buckets:
                name = f"{name} - External AWS"
            s3_buckets[name] = external_s3["path"]

        for internal_s3 in self.openapi.s3:
            name = internal_s3["map_settings"].get("name", "")
            if len(name) == 0 or name in s3_buckets:
                name = internal_s3["map_settings"]["id"]
            if name in s3_buckets:
                name = f"{name} - Internal AWS"
            s3_buckets[name] = internal_s3["map_settings"]["path"]

        return s3_buckets

    def listdir(
        self,
        path: str,
        ignore_hidden: bool = True,
        get_details: bool = False,
    ) -> Union[List[Dict[str, Union[str, int, dict]]], List[str]]:
        """
        List all files and folders with path in domain server

        Parameters
        ----------
        path: ``str``
            path of folder to list
        ignore_hidden: ``bool``, default: True
            Ignore hidden files/folders or not
        get_details: ``bool``, default: False
            Get details information or not

        Returns
        -------
        results: ``Union[List[Dict[str, Union[str, int, dict]]], List[str]]``
            Folders and files with their information
        """
        dir_elements = self.openapi.list_dir(
            path, ignore_hidden=ignore_hidden
        )[ConnectorKeys.ENTITIES.value]
        if get_details:
            return dir_elements
        return [element[ConnectorKeys.NAME.value] for element in dir_elements]

    def parse_data_information(self, data_name: str, technology: str, data_path: str) -> dict:
        """
        Parse information of data to valid format for submission

        Parameters
        ----------
        data_name: ``str``
            Name of spatial data
        technology: ``str``
            Technology of spatial data
        data_path: ``str``
            Path to spatial data

        Returns
        -------
        submission_data: ``dict``
            Auto-detect information for submission
        """
        return self.pyapi.parse_data_information(data_name, technology, data_path)

    def parse_multiple_samples_information(
        self,
        technology: str,
        data_path: str,
        sample_name_mapping: dict = {},
        data_name_mapping: dict = {},
    ) -> List[dict]:
        """
        Parse information of multiple samples to valid format for submission

        Parameters
        ----------
        technology: ``str``
            Technology of spatial data
        data_path: ``str``
            Path to spatial data

        Returns
        -------
        submission_samples: ``dict``
            Auto-detect information for submission
        """
        results: List[dict] = []
        for folder in self.listdir(data_path, get_details=True):
            if folder[ConnectorKeys.TYPE.value] != ConnectorKeys.DIRECTORY.value:
                continue
            try:
                data = self.pyapi.parse_data_information(
                    data_name_mapping.get(folder[ConnectorKeys.NAME.value], folder[ConnectorKeys.NAME.value]),
                    technology,
                    os.path.join(data_path, folder[ConnectorKeys.NAME.value]),
                )
                results.append({
                    ConnectorKeys.SAMPLE_NAME.value: sample_name_mapping.get(
                        folder[ConnectorKeys.NAME.value], folder[ConnectorKeys.NAME.value]
                    ),
                    ConnectorKeys.DATA.value: data,
                })
            except Exception as e:
                print(f"Fail to parse data in {folder[ConnectorKeys.NAME.value]}: {e}")

        return results

    def list_study(
        self, group: str, species: str, **kwargs
    ) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        List reachable studies

        Parameters
        ----------
        group: ``str``
            Group of studies
        species: ``str``
            Species of studies

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            List of studies and their information

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.list_study(self.groups[group], species, **kwargs)["list"]

    def get_study_detail(self, study_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        Get details information of study

        Parameters
        ----------
        study_id: ``str``
            Id of study

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of study

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_study_detail(study_id, **kwargs)

    def list_sample(self, study_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        List samples in a study

        Parameters
        ----------
        study_id: ``str``
            Id of study

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            List of samples and their information

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.list_sample(study_id, **kwargs)["list"]

    def get_sample_detail(self, sample_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        Get details information of sample

        Parameters
        ----------
        sample_id: ``str``
            Id of sample

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of sample

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_sample_detail(sample_id, **kwargs)

    def get_sample_data_detail(self, data_id: str, **kwargs) -> Dict[str, Union[str, dict, List[dict]]]:
        """
        Get details information of sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data

        Returns
        -------
        results: ``Dict[str, Union[str, dict, List[dict]]]``
            Information of data

        Keyword Arguments
        -----------------
        limit: ``int``, default: 50
            Limit of responsing data
        offset: ``int``, default: 50
            Starting point of responsing data
        """
        return self.openapi.get_sample_data_detail(data_id, **kwargs)

    def get_sample_data_elements(self, data_id: str) -> Dict[str, List[str]]:
        """
        Get elements of sample data

        Parameters
        ----------
        data_id: ``str``
            Id of data

        Returns
        -------
        results: ``Dict[str, List[str]]``
            Elements of data
        """
        results: Dict[str, Dict[str, dict]] =  \
            self.openapi.get_sample_data_detail(data_id).get("map_submit_result", None)
        if results is None:
            return {}
        elements: Dict[str, List[str]] = {}
        for value in results.values():
            if not isinstance(value, dict):
                continue
            for k, v in value.items():
                if k not in elements:
                    elements[k] = []
                elements[k].extend([v] if isinstance(v, str) else v.values())
        return elements

    def add_sample(
        self,
        study_id: str,
        sample_name: str,
        sample_data: List[dict] = [],
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Add a sample to a existed study

        Parameters
        ----------
        study_id: ``str``
            Id of study
        sample_name: ``str``
            Sample name
        sample_data: ``List[dict]``, default: []
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """

        return self.openapi.create_sample(study_id, sample_name, sample_data)

    def add_sample_data(
        self,
        study_id: str,
        sample_id: str,
        sample_data: List[dict],
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Add data to a existed sample

        Parameters
        ----------
        study_id: ``str``
            Id of sample's root study
        sample_id: ``str``
            Id of sample
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """
        return self.openapi.add_sample_data(study_id, sample_id, sample_data)

    def submit(
        self,
        group: str,
        species: str,
        title: str,
        sample_name: str,
        sample_data: List[dict],
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        group: ``str``
            Group of study
        species: ``str``
            Species of data in study
        title: ``str``
            Title of study
        sample_name: ``str``
            Sample name
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_data_information`` function

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Submission information
        """
        study_id = self.openapi.create_study(self.groups[group], species, title)[ConnectorKeys.STUDY_ID.value]
        return self.add_sample(study_id, sample_name, sample_data)

    def submit_multiple_samples(
        self,
        group: str,
        species: str,
        title: str,
        sample_data: List[dict],
    ) -> List[Dict[str, Union[str, List[dict]]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        group: ``str``
            Group of study
        species: ``str``
            Species of data in study
        title: ``str``
            Title of study
        sample_data: ``List[dict]``
            List of data in sample, each data is result of ``parse_multiple_samples_information`` function

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        study_id = self.openapi.create_study(self.groups[group], species, title)[ConnectorKeys.STUDY_ID.value]

        results = []
        for sample in sample_data:
            sample_name = sample[ConnectorKeys.SAMPLE_NAME.value]
            data = sample[ConnectorKeys.DATA.value]
            results.append(self.add_sample(study_id, sample_name, data))

        return results

    def add_sample_data_element(
        self,
        title: str,
        study_id: str,
        sample_id: str,
        data_id: str,
        adding_types: List[str],
        paths: Dict[str, str] = {},
        args: Dict[str, str] = {},
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Create new study and submit the first sample.

        Parameters
        ----------
        title: ``str``
            Title of elements
        study_id: ``str``
            Study ID
        sample_id: ``str``
            Sample ID
        data_id: ``str``
            Data ID
        adding_types: `List[str]`
            Type of element to adding, defined in [
                `spatialx_connector.ExtendImagesSubmission`,
                `spatialx_connector.ExtendSegmentationSubmission`,
                `spatialx_connector.ExtendTrasncriptsSubmission`,
                `spatialx_connector.ExtendExpressionSubmission`,
            ]
        paths: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and their paths.
        args: `Dict[str, str]`
            Mapping of elements defined in `spatialx_connector.SubmissionElementKeys` and values.

        Returns
        -------
        results: ``List[Dict[str, Union[str, List[dict]]]]``
            Submission information
        """
        return self.openapi.add_sample_data_element(
            title,
            study_id,
            sample_id,
            data_id,
            identities=adding_types,
            files=[
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in paths.items()
            ],
            args=[
                {
                    ConnectorKeys.KEY.value: key,
                    ConnectorKeys.VALUE.value: value,
                }
                for key, value in args.items()
            ] + [
                {
                    ConnectorKeys.KEY.value: "ignore_technology_elements",  # Ignore default types of technologies.
                    ConnectorKeys.VALUE.value: True,
                }
            ],
        )

    def upload_file(
        self,
        file_path: str,
        server_folder_name: str = "",
        upload_id: str = "",
        is_chunk: bool = False,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        upload a small file

        Parameters
        ----------
        file_path: ``str``
            File location
        server_folder_name: ``str``
            Folder location in spatialx server
        upload_id: ``str``
            Upload ID

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Upload information
        """
        return self.openapi.upload_file(
            file_path=file_path,
            folder_name=server_folder_name,
            upload_id=upload_id,
            is_chunk=is_chunk,
        )

    def upload_big_file(
        self,
        file_path: str,
        chunk_size: int = 0,
        debug_mode: bool = False,
        server_folder_name: str = "",
        chunk_resp: dict = {},
        move_to_parent: bool = True,
    ) -> Dict[str, Union[str, List[dict]]]:
        """
        Upload a big file

        Parameters
        ----------
        file_path: ``str``
            File location
        chunk_size: ``int``
            Chunk size (bytes), 0: auto
        debug_mode: ``bool``
            Debug mode
        server_folder_name: ``str``
            Folder location in spatialx server

        Returns
        -------
        results: ``Dict[str, Union[str, List[dict]]]``
            Upload information
        """
        if not os.path.isfile(file_path):
            raise Exception(f"Invalid file: {file_path}")

        file_size = os.stat(os.path.abspath(file_path)).st_size
        upload_id = ""
        resp = chunk_resp
        if ConnectorKeys.UNIQUE_ID.value in resp:
            upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

        # Direct upload if small file
        if file_size < ConnectorKeys.UPLOAD_CHUNK_SMALL_SIZE.value:
            if ConnectorKeys.UNIQUE_ID.value in resp:
                upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

            return self.upload_file(
                file_path=file_path,
                server_folder_name=server_folder_name,
                upload_id=upload_id,
                is_chunk=True,
            )

        file_name = Path(file_path).name
        item_chunk_size = get_chunk_size(chunk_size, file_size)

        if (len(resp.keys()) == 0) or (len(upload_id) == 0):
            resp = self.openapi.upload_chunk_start(
                folder_name=server_folder_name,
                parent_is_file=2,
            )

            if ConnectorKeys.UNIQUE_ID.value in resp:
                upload_id = resp[ConnectorKeys.UNIQUE_ID.value]

        file = open(file_path, "rb")
        file.seek(0, 0)
        sending_index = 0
        offset_size = 0
        progress_bar = None
        if debug_mode:
            progress_bar = tqdm(total=file_size, unit="B", unit_scale=True)

        while True:
            data = file.read(item_chunk_size)
            if not data:
                break

            offset_size = offset_size + item_chunk_size
            offset_size = min(file_size, offset_size)

            if debug_mode:
                format_print(f"Upload {file_path}, chunk index : {sending_index + 1} ...")

            self.openapi.upload_chunk_process(
                chunk_size=item_chunk_size,
                file_size=file_size,
                offset=offset_size,
                file_name=file_name,
                folder_name=server_folder_name,
                upload_id=upload_id,
                path=resp[ConnectorKeys.ROOT_FOLDER.value],
                sending_index=sending_index,
                parent_is_file=2,
                file_data=data,
            )

            if debug_mode:
                if progress_bar is not None:
                    progress_bar.update(len(data))

            sending_index = sending_index + 1

        total_index = sending_index
        file.close()

        resp2 = self.openapi.upload_chunk_merge(
            total_chunk=total_index,
            file_name=file_name,
            folder_name=server_folder_name,
            upload_id=upload_id,
            path=resp[ConnectorKeys.ROOT_FOLDER.value],
            parent_is_file=2,
            move_to_parent=move_to_parent,
        )

        if move_to_parent:
            return resp2
        return resp

    def upload_folder(
        self,
        dir_path: str,
        folder_path: Optional[str] = None,
        chunk_size: int = 0,
        debug_mode: bool = False,
        server_folder_name: str = "",
        chunk_resp: dict = {},
    ) -> bool:
        """
        Upload folder as: zarr

        Parameters
        ----------
        dir_path: ``str``
            Folder location
        chunk_size: ``int``
            Chunk size (bytes), 0: auto
        debug_mode: ``bool``
            Debug mode
        server_folder_name: ``str``
            Folder location in spatialx server
        """
        if not os.path.isdir(dir_path):
            raise Exception(f"Invalid directory: {dir_path}")

        root_folder_path = ""
        if folder_path is None:
            folder_path = server_folder_name + os.path.basename(dir_path)
            root_folder_path = str(folder_path)

        src_path = Path(dir_path)
        resp = chunk_resp

        for src_child in src_path.iterdir():
            if src_child.is_dir():
                folder_path = os.path.join(folder_path, src_child.stem)
                dst_child = os.path.join(dir_path, src_child.stem)
                self.upload_folder(
                    dir_path=dst_child, folder_path=folder_path,
                    chunk_size=chunk_size, debug_mode=debug_mode,
                    server_folder_name=server_folder_name,
                    chunk_resp=resp,
                )
            else:
                if src_child.is_symlink():
                    continue

                dst_child = os.path.join(dir_path, src_child.name)
                resp = self.upload_big_file(
                    file_path=dst_child,
                    chunk_size=chunk_size,
                    debug_mode=debug_mode,
                    server_folder_name=folder_path,
                    chunk_resp=resp,
                    move_to_parent=False,
                )

        return self.openapi.upload_folder_finish(
            root_folder_path,
            resp[ConnectorKeys.UNIQUE_ID.value],
        )

    def list_lens_bulk_studies(self, host: str, token: str, group: str, species: str):
        from bioturing_connector.lens_bulk_connector import LensBulkConnector

        connector = LensBulkConnector(host=host, token=token, ssl=True)
        connector.test_connection()
        if group == DefaultGroup.PERSONAL_WORKSPACE.value:
            group_id = DefaultGroup.LENS_GROUP_ID_PERSONAL_WORKSPACE.value
        elif group == DefaultGroup.ALL_MEMBERS.value:
            group_id = DefaultGroup.LENS_GROUP_ID_ALL_MEMBERS.value
        else:
            group_id = self.groups.get(group, group)

        studies_info = connector.get_all_studies_info_in_group(group_id=group_id, species=species)
        studies_info = [
            {
                **info,
                ConnectorKeys.SPECIES.value: species,
                ConnectorKeys.GROUP_ID.value: self.groups.get(
                    group, DefaultGroup.PERSONAL_WORKSPACE.value
                ),
            }
            for info in studies_info
        ]
        return studies_info

    def list_lens_sc_studies(self, host: str, token: str, group: str, species: str):
        from bioturing_connector.lens_sc_connector import LensSCConnector

        connector = LensSCConnector(host=host, token=token, ssl=True)
        connector.test_connection()
        if group == DefaultGroup.PERSONAL_WORKSPACE.value:
            group_id = DefaultGroup.LENS_GROUP_ID_PERSONAL_WORKSPACE.value
        elif group == DefaultGroup.ALL_MEMBERS.value:
            group_id = DefaultGroup.LENS_GROUP_ID_ALL_MEMBERS.value
        else:
            group_id = self.groups[group]

        studies_info = connector.get_all_studies_info_in_group(group_id=group_id, species=species)
        studies_info = [
            {
                **info,
                ConnectorKeys.SPECIES.value: species,
                ConnectorKeys.GROUP_ID.value: self.groups[group],
            }
            for info in studies_info
        ]
        return studies_info

    def convert_data_from_lens(self, study_info: Union[dict, List[dict]]):
        if isinstance(study_info, dict):
            study_info = [study_info]

        for info in study_info:
            self.openapi.convert_from_lens(
                f"ST-{info['id']}",
                info["title"],
                info[ConnectorKeys.GROUP_ID.value],
                info[ConnectorKeys.SPECIES.value],
                info["id"],
            )



from setuptools import setup, find_packages
from pathlib import Path


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

BASE_PACKAGE = "spatialx_connector"
BASE_IMPORT = "spatialx_connector"


def _install_requires():
    return [
        "pydantic==2.7.2",
        "requests==2.32.3",
        "tqdm==4.66.4",
    ]


setup(
    name=BASE_PACKAGE,
    version="0.1.6",
    author="BioTuring",
    author_email="support@bioturing.com",
    url="https://alpha.bioturing.com",
    description="BioTuring SpatialX Connector",
    package_dir={BASE_IMPORT: "spatialx_connector"},
    packages=[BASE_IMPORT, *find_packages()],
    zip_safe=False,
    install_requires=_install_requires(),
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from requests.adapters import Retry
from requests.adapters import HTTPAdapter
import requests
from ._constants import ConnectorKeys


TAB_STRING: str = " " * 4


class RequestMode:
    POST: str = "post"
    PUT: str = "put"
    GET: str = "get"


def create_requests_session():
    session = requests.Session()
    retries = Retry(
        total=5,
        backoff_factor=0.1
    )
    session.mount('http://', HTTPAdapter(max_retries=retries))
    session.mount('https://', HTTPAdapter(max_retries=retries))
    return session


def request(
    url: str,
    params={},
    data={},
    json={},
    headers={},
    files={},
    mode: str = RequestMode.POST,
    verify: bool = False,
):
    session = create_requests_session()
    r: requests.Response = getattr(session, mode)(
        url, params=params, json=json, data=data,
        headers=headers, verify=verify, files=files,
        timeout=1800
    )

    if r.status_code >= 400:
        raise Exception(
            f"Call request to {url} failed with status code {r.status_code}, response {r.text}"
        )
    try:
        response = r.json()
    except:
        response = r.text
    r.close()
    return response


def parse_to_str(data, tab_level: int = 0) -> str:
    if isinstance(data, list):
        return list_to_str(data, tab_level)
    if isinstance(data, tuple):
        return tuple_to_str(data, tab_level)
    if isinstance(data, dict):
        return dict_to_str(data, tab_level)
    return f"{data}"


def list_to_str(data: list, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "[""]"
    return "[\n" + "".join([
        TAB_STRING * (tab_level + 1) +
        f"{parse_to_str(value, tab_level + 1)}\n" for value in data
    ]) + TAB_STRING * tab_level + "]"


def tuple_to_str(data: list, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "("")"
    return "(\n" + "".join([
        TAB_STRING * (tab_level + 1) +
        f"{parse_to_str(value, tab_level + 1)}\n" for value in data
    ]) + TAB_STRING * tab_level + ")"


def dict_to_str(data: dict, tab_level: int = 0) -> str:
    if len(data) == 0:
        return "{""}"
    return "{\n" + "".join([
        f"{TAB_STRING * (tab_level + 1)}{key}: "
        f"{parse_to_str(value, tab_level + 1)}\n"
        for key, value in data.items()
    ]) + TAB_STRING * tab_level + "}"


def format_print(data):
    print(parse_to_str(data))


def get_chunk_size(chunk_size: int, file_size: int) -> int:
    if chunk_size > 0:
        if chunk_size > ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value:
            chunk_size = ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value
        return chunk_size

    if file_size < 15*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_SMALL_SIZE.value
    elif file_size < 100*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_MEDIUM_SIZE.value
    elif file_size < 1024*1024*1024:
        return ConnectorKeys.UPLOAD_CHUNK_NORMAL_SIZE.value

    return ConnectorKeys.UPLOAD_CHUNK_LARGE_SIZE.value



from typing import Union, List, Dict
import os
from urllib.parse import urlparse, parse_qs

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from . import _constants as constants
from ._utils import RequestMode
from ._utils import request
from ._utils import parse_to_str


class Connector:
    def __init__(
        self,
        domain: str,
        token: str,
        verify_ssl: bool = False,
    ):
        self._token = token
        self._verify_ssl = verify_ssl

        link = urlparse(domain)
        if len(link.netloc) == 0:
            raise Exception("invalid domain: {}".format(domain))

        params = dict(parse_qs(link.query))
        params = {k: v[0] for k, v in params.items()}
        self.params = params
        self._domain = "{}://{}{}".format(link.scheme, link.netloc, link.path)

    def request(
            self,
            url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ):
        for k, v in self.params.items():
            params[k] = v

        res = request(
            url=url.replace(os.sep, "/"),
            params=params,
            json=json,
            data=data,
            files=files,
            headers={constants.TOKEN_KEY: self._token},
            mode=mode,
            verify=self._verify_ssl
        )
        if res[constants.ConnectorKeys.STATUS.value] != 0:
            raise Exception(parse_to_str(res))

        return res[constants.ConnectorKeys.MESSAGE.value]

    def post_pyapi_request(
            self, url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ) -> Union[BaseModel, str]:
        return self.request(
            os.path.join(self._domain, constants.V1_API, url),
            params=params, json=json, data=data,
            mode=mode, files=files,
        )

    def post_openapi_request(
            self, url: str,
            params: dict = {},
            json: dict = {},
            data: dict = {},
            mode: str = RequestMode.POST,
            files: dict = {},
    ) -> dict:
        return self.request(
            os.path.join(self._domain, constants.OPEN_API, url),
            params=params, json=json, data=data,
            mode=mode, files=files,
        )


class PyAPI(Connector):
    def parse_data_information(self, name: str, technology: str, data_path: str) -> dict:
        return self.post_pyapi_request(
            url=constants.PARSE_DATA_URL,
            json={
                constants.ConnectorKeys.DATA_NAME.value: name,
                constants.ConnectorKeys.TECHNOLOGY.value: technology,
                constants.ConnectorKeys.DATA_PATH.value: data_path,
            }
        )

    def get_sample_info(
        self,
        study_path: str,
        sample_id: str,
    ):
        return self.post_pyapi_request(
            url=constants.SAMPLE_INFO_URL,
            json={
                constants.ConnectorKeys.STUDY_PATH.value: study_path,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
            },
        )


class OpenAPI(Connector):
    @property
    def info(self):
        return self.post_openapi_request(
            url=constants.INFO_URL,
            mode=RequestMode.GET,
        )

    @property
    def mounts(self):
        return self.post_openapi_request(
            url=constants.EXTERNAL_MOUNT_URL,
            mode=RequestMode.GET,
        )

    def list_s3(self, offset: int = 0, limit: int = 100):
        return self.post_openapi_request(
            url=constants.LIST_S3,
            mode=RequestMode.POST,
            data={
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            }
        )

    @property
    def s3(self):
        return self.list_s3()

    @property
    def groups(self):
        return self.post_openapi_request(
            url=constants.GROUPS_URL,
            mode=RequestMode.GET,
        )

    def list_dir(self, path: str, ignore_hidden: bool = True):
        return self.post_openapi_request(
            constants.LIST_URL,
            data={
                constants.ConnectorKeys.PATH.value: path,
                constants.ConnectorKeys.IGNORE_HIDDEN.value: ignore_hidden,
            }
        )

    def create_study(
        self,
        group_id: str,
        species: str,
        title: str,
        create_type: int = constants.StudyType.SPATIAL_STUDY_TYPE_NUMBER.value,
    ):
        return self.post_openapi_request(
            url=constants.CREATE_STUDY_URL,
            json={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.TITLE.value: title,
                constants.ConnectorKeys.TYPE.value: create_type,
            }
        )

    def list_study(
        self,
        group_id: str,
        species: str,
        limit: int = 50,
        offset: int = 0,
        active: int = constants.StudyStatus.PROCESSING_STATUS.value,
        compare: int = constants.StudyFilter.NOT_LARGER.value,
    ):
        return self.post_openapi_request(
            url=constants.LIST_STUDY_URL,
            data={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.ACTIVE.value: active,
                constants.ConnectorKeys.COMPARE.value: compare,
            }
        )

    def get_study_detail(self, study_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_STUDY_URL,
            params={
                constants.ConnectorKeys.KEY.value: study_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def create_sample(
        self,
        study_id: str,
        name: str,
        data: List[dict],
    ):
        return self.post_openapi_request(
            url=constants.CREATE_SAMPLE_URL,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.NAME.value: name,
                constants.ConnectorKeys.DATA.value: data,
            }
        )

    def add_sample_data(
        self,
        study_id: str,
        sample_id: str,
        data: List[dict],
    ):
        return self.post_openapi_request(
            url=constants.ADD_SAMPLE_DATA_URL,
            json={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.DATA.value: data,
            }
        )

    def add_sample_data_element(
        self,
        title: str,
        study_id: str,
        sample_id: str,
        data_id: str,
        identities: List[str] = [],
        files: List[Dict[str, str]] = [],
        folders: List[Dict[str, str]] = [],
        args: List[Dict[str, str]] = [],
    ):
        return self.post_openapi_request(
            url=constants.ADD_URL,
            json={
                constants.ConnectorKeys.TITLE.value: title,
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.SAMPLE_ID.value: sample_id,
                constants.ConnectorKeys.DATA_ID.value: data_id,
                constants.ConnectorKeys.FILES.value: files,
                constants.ConnectorKeys.FOLDERS.value: folders,
                constants.ConnectorKeys.ARGS.value: args,
                constants.ConnectorKeys.IDENTITIES.value: identities,
            },
        )

    def list_sample(
        self,
        study_id: str,
        limit: int = 50,
        offset: int = 0,
        need_data: bool = False,
    ):
        return self.post_openapi_request(
            url=constants.LIST_SAMPLE_URL,
            params={
                constants.ConnectorKeys.KEY.value: study_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.NEED_DATA.value: need_data,
            },
            mode=RequestMode.GET,
        )

    def get_sample_detail(self, sample_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_SAMPLE_URL,
            params={
                constants.ConnectorKeys.KEY.value: sample_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def get_sample_data_detail(self, data_id: str, limit: int = 50, offset: int = 0):
        return self.post_openapi_request(
            url=constants.DETAIL_SAMPLE_DATA_URL,
            params={
                constants.ConnectorKeys.KEY.value: data_id,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
            },
            mode=RequestMode.GET,
        )

    def list_public_study(
        self,
        group_id: str,
        species: str,
        limit: int = 50,
        offset: int = 0,
        active: int = constants.StudyStatus.PROCESSING_STATUS.value,
    ):
        return self.post_openapi_request(
            url=constants.LIST_PUBLIC_STUDY_URL,
            json={
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LIMIT.value: limit,
                constants.ConnectorKeys.OFFSET.value: offset,
                constants.ConnectorKeys.ACTIVE.value: active,
            }
        )

    def upload_file(
        self, file_path: str,
        folder_name: str, upload_id: str,
        is_chunk: bool,
    ):
        file = open(file_path, "rb")
        resp = self.post_openapi_request(
            url=constants.UPLOAD_FILE_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_IS_CHUNK.value: is_chunk,
            },
            files={
                constants.ConnectorKeys.UPLOAD_FILE_DATA.value: file,
            },
        )
        file.close()
        return resp

    def upload_chunk_start(self, folder_name: str, parent_is_file: int):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_START_URL,
            json={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
            }
        )

    def upload_chunk_process(
        self,
        chunk_size: int,
        file_size: int,
        offset: int,
        file_name: str,
        folder_name: str,
        upload_id: str,
        path: str,
        sending_index: int,
        parent_is_file: int,
        file_data: list[str],
    ):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_PROCESS_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
                constants.ConnectorKeys.UPLOAD_CHUNK_SIZE.value: chunk_size,
                constants.ConnectorKeys.UPLOAD_FILE_SIZE.value: file_size,
                constants.ConnectorKeys.UPLOAD_OFFSET.value: offset,
                constants.ConnectorKeys.UPLOAD_FILE_NAME.value: file_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_PATH.value: path,
                constants.ConnectorKeys.UPLOAD_SENDING_INDEX.value: sending_index,
            },
            files={
                constants.ConnectorKeys.UPLOAD_FILE_DATA.value: file_data,
            }
        )

    def upload_chunk_merge(
        self,
        total_chunk: int,
        file_name: str,
        folder_name: str,
        upload_id: str,
        path: str,
        parent_is_file: int,
        move_to_parent: bool,
    ):
        return self.post_openapi_request(
            url=constants.UPLOAD_CHUNK_MERGE_URL,
            json={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_PARENT_IS_FILE.value: parent_is_file,
                constants.ConnectorKeys.UPLOAD_TOTAL_CHUNK.value: total_chunk,
                constants.ConnectorKeys.UPLOAD_FILE_NAME.value: file_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
                constants.ConnectorKeys.UPLOAD_PATH.value: path,
                constants.ConnectorKeys.UPLOAD_MOVE_TO_PARENT.value: move_to_parent,
            }
        )

    def upload_folder_finish(self, folder_name: str, upload_id: str):
        return self.post_openapi_request(
            url=constants.UPLOAD_FOLDER_FINISH_URL,
            data={
                constants.ConnectorKeys.UPLOAD_FOLDER_NAME.value: folder_name,
                constants.ConnectorKeys.UPLOAD_UNIQUE_ID.value: upload_id,
            },
        )

    def convert_from_lens(
        self, study_id: str, study_name: str,
        group_id: str, species: str, lens_data_path: str
    ):
        return self.post_openapi_request(
            url=constants.STUDY_CONVERT_FROM_LENS_URL,
            data={
                constants.ConnectorKeys.STUDY_ID.value: study_id,
                constants.ConnectorKeys.TITLE.value: study_name,
                constants.ConnectorKeys.GROUP_ID.value: group_id,
                constants.ConnectorKeys.SPECIES.value: species,
                constants.ConnectorKeys.LENS_DATA_PATH.value: lens_data_path,
            }
        )



from ._utils import format_print
from ._constants import (
    Species,
    Technologies,
    DefaultGroup,
    ConnectorKeys,
    SubmissionElementKeys,
    ExtendImagesSubmission,
    ExtendSegmentationSubmission,
    ExtendTrasncriptsSubmission,
    ExtendExpressionSubmission,
)
from ._connector import SpatialXConnector


__ALL__ = [
    Species,
    Technologies,
    DefaultGroup,
    ConnectorKeys,
    SubmissionElementKeys,
    ExtendImagesSubmission,
    ExtendSegmentationSubmission,
    ExtendTrasncriptsSubmission,
    ExtendExpressionSubmission,
    SpatialXConnector,
    format_print,
]



from enum import Enum


V1_API: str = "pyapi/v1"
OPEN_API: str = "openapi"
TOKEN_KEY: str = "SpatialX-Token"


# PyAPI
SUBMIT_URL: str = "submission/submit_data"
ADD_URL: str = "data_extend/add"
PARSE_DATA_URL: str = "submission/parse_data_information"

SAMPLE_INFO_URL: str = "connector/get_sample_info"

# OpenAPI
INFO_URL: str = "account/info"
GROUPS_URL: str = "account/groups"
LIST_URL: str = "directory/entities"
EXTERNAL_MOUNT_URL: str = "mount/external_listing"
LIST_S3: str = "cloud/setting_list"

CREATE_STUDY_URL: str = "study/create"
LIST_STUDY_URL: str = "study/list"
LIST_PUBLIC_STUDY_URL: str = "study/list_public"
DETAIL_STUDY_URL: str = "study/detail"

CREATE_SAMPLE_URL: str = "sample/create"
LIST_SAMPLE_URL: str = "sample/list"
DETAIL_SAMPLE_URL: str = "sample/detail"

DETAIL_SAMPLE_DATA_URL: str = "data/detail"

ADD_SAMPLE_DATA_URL: str = "data/create"
ADD_SAMPLE_DATA_ELEMENT_URL: str = "data/add_element"

UPLOAD_FILE_URL: str = "upload/simple"
UPLOAD_CHUNK_START_URL: str = "upload/chunk/start"
UPLOAD_CHUNK_PROCESS_URL: str = "upload/chunk/process"
UPLOAD_CHUNK_MERGE_URL: str = "upload/chunk/merge"
UPLOAD_CHUNK_FINISH_URL: str = "upload/chunk/finish"
UPLOAD_FOLDER_FINISH_URL: str = "upload/folder_finish"


# Convert from Lens
STUDY_CONVERT_FROM_LENS_URL: str = "study/convert_from_lens"

# Analysis
CREATE_ANALYSIS_URL: str = "analysis/create"
GET_ANNOTATED_ELEMENTS: str = "metadata/get_annotated_element"


# Types
class Technologies(Enum):
    COSMX_VER1 = "COSMX_VER1"
    COSMX_VER2 = "COSMX_VER2"
    MERSCOPE_VER1 = "MERSCOPE_VER1"
    MERSCOPE_VER2 = "MERSCOPE_VER2"
    XENIUM = "XENIUM"
    XENIUM_HE = "XENIUM_HE"
    VISIUM = "VISIUM"
    VISIUM_HD = "VISIUM_HD"
    SPATIALDATA_ZARR = "SPATIALDATA_ZARR"
    SPATIALDATA_H5AD = "SPATIALDATA_H5AD"


class Species(Enum):
    HUMAN = "human"
    MOUSE = "mouse"
    OTHERS = "others"
    NON_HUMAN_PRIMATE = "nonHumanPrimate"


class StudyType(Enum):
    SINGLECELL_STUDY_TYPE_NUMBER = 0
    SPATIAL_STUDY_TYPE_NUMBER = 1


class StudyStatus(Enum):
    CREATED_STATUS = 0
    SUCCESS_STATUS = 1
    PROCESSING_STATUS = 2
    DELETE_STATUS = 3


class StudyFilter(Enum):
    EQUAL = 0
    NOT_LARGER = 1
    LARGER = 2


class DefaultGroup(Enum):
    PERSONAL_WORKSPACE = "Personal workspace"
    ALL_MEMBERS = "All members"

    LENS_GROUP_ID_PERSONAL_WORKSPACE = "personal"
    LENS_GROUP_ID_ALL_MEMBERS = "all_members"


class ExtendImagesSubmission(Enum):
    TIFFFILE = "IMAGES_TIFFFILE"
    TIFFFILE_3D = "IMAGES_TIFFFILE_3D"
    EXISTED_IMAGES = "IMAGES_FROM_EXISTED"
    EXISTED_PROTEIN_IMAGES = "IMAGES_PROTEIN_FROM_EXISTED"


class ExtendSegmentationSubmission(Enum):
    PARQUET = "SEGMENTATION_PARQUET"
    GEOJSON = "SEGMENTATION_GEOJSON"
    FEATHER = "SEGMENTATION_FEATHER"
    HALO = "SEGMENTATION_HALO"

    FROM_EXISTED = "SEGMENTATION_FROM_EXISTED"
    FROM_IMAGES = "SEGMENTATION_FROM_PROTEIN_IMAGES"


class ExtendTrasncriptsSubmission(Enum):
    DATAFRAME = "TRANSCRIPTS_DATAFRAME"


class ExtendExpressionSubmission(Enum):
    IMPORT_ANNDATA = "IMPORT_ANNDATA"


class SubmissionElementKeys(Enum):
    IMAGES = "images"
    PROTEIN_IMAGES = "protein_images"
    SEGMENTATION = "segmentation"
    TRANSCRIPTS = "transcripts"
    CELL_CENTERS = "cell_centers"
    EXPRESSION = "annotated_data"
    ALIGNMENT = "alignment"

    IMAGES_ID = "images_id"
    SEGMENTATION_ID = "segmentation_id"
    SPATIAL_ID = "spatial_id"
    NUCLEI_CHANNELS = "nuclei_channels"
    MEMBRANE_CHANNELS = "membrane_channels"


class ConnectorKeys(Enum):
    INFORMATION_FIELDS = ["email", "sub_dir", "name", "app_base_url", "routing_table"]

    # Response keys
    ENTITY = "entity"
    ENTITIES = "entities"
    MESSAGE = "message"
    STATUS = "status"
    STUDY = "study"
    SAMPLE = "sample"
    UNIQUE_ID = "unique_id"
    ROOT_FOLDER = "root_folder"

    FILE = "file"
    DIRECTORY = "directory"

    FILES = "files"
    FOLDERS = "folders"
    ARGS = "args"
    IDENTITIES = "identities"

    MAP_SUBMIT_RESULT = "map_submit_result"
    ANNOTATED_DATA = "annotated_data"
    TRANSCRIPTS = "transcripts"
    SPOT = "spot"

    # Parameter keys
    OBS = "obs"
    OBSM = "obsm"
    BARCODES: str = "barcodes"
    GENES: str = "genes"

    STUDY_PATH = "study_path"
    STUDY_ID = "study_id"
    GROUP_ID = "group_id"
    SPECIES = "species"
    LIMIT = "limit"
    OFFSET = "offset"
    ACTIVE = "active"
    COMPARE = "compare"
    NAME = "name"
    DATA = "data"
    DATA_NAME = "data_name"
    SAMPLE_NAME = "sample_name"
    TITLE = "title"
    KEY = "key"
    VALUE = "value"
    TYPE = "type"
    PATH = "path"
    DATA_PATH = "data_path"
    IGNORE_HIDDEN = "ignore_hidden"
    TECHNOLOGY = "technology"
    SAMPLE_ID = "sample_id"
    SAMPLE_DATA = "sample_data"
    SUBMIT_ID = "submit_id"
    SUBMISSION_NAME = "submission_name"
    SUBMISSION_TYPE = "submission_type"
    SUBMISSION_INFO = "submission_info"
    NEED_DATA = "need_data"
    DATA_ID = "data_id"
    TABLE_ID = "table_id"
    LENS_DATA_PATH = "lens_data_path"
    GROUPS = "groups"
    DEFAULT = "default"
    PARAMS = "params"
    DISPLAY_PARAMS = "display_params"
    SUB_TYPE = "sub_type"
    GROUP_TYPE = "group_type"
    DESCRIPTION = "description"
    ELEMENT = "element"

    # Parameter upload keys
    UPLOAD_PARENT_IS_FILE = "parent_is_file"
    UPLOAD_CHUNK_SIZE = "chunk_size"
    UPLOAD_FILE_SIZE = "file_size"
    UPLOAD_OFFSET = "offset"
    UPLOAD_FILE_NAME = "name"
    UPLOAD_FOLDER_NAME = "folder_name"
    UPLOAD_UNIQUE_ID = "unique_id"
    UPLOAD_PATH = "path"
    UPLOAD_MOVE_TO_PARENT = "move_to_parent"
    UPLOAD_SENDING_INDEX = "sending_index"
    UPLOAD_FILE_DATA = "file"
    UPLOAD_TOTAL_CHUNK = "total"
    UPLOAD_IS_CHUNK = "is_chunk"
    UPLOAD_CHUNK_SMALL_SIZE = 1024 * 1024
    UPLOAD_CHUNK_MEDIUM_SIZE = 16 * 1024 * 1024
    UPLOAD_CHUNK_NORMAL_SIZE = 50 * 1024 * 1024
    UPLOAD_CHUNK_LARGE_SIZE = 100 * 1024 * 1024


class NormalizeMethod(Enum):
    RAW = "raw"
    LOG1P_NORMALIZE = "log1p-normalized"
    SQRT_NORMALIZE = "sqrt-normalized"
